'''
    Created on Sep 29, 2011

    @author: ksahlin

    This file is part of BESST.
    
    BESST is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    BESST is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with BESST.  If not, see <http://www.gnu.org/licenses/>.
    '''

import sys
from collections import defaultdict
import time
import copy

import networkx as nx
from networkx import algorithms
import multiprocessing

import Contig, Scaffold, Parameter
import GenerateOutput as GO
from mathstats.normaldist.truncatedskewed import param_est as GC
from mathstats.normaldist import normal
import ExtendLargeScaffolds as ELS
import haplotypes as HR
import plots
import order_contigs


def constant_large():
    return 2 ** 32

def constant_small():
    return -1


def Algorithm(G, G_prime, Contigs, small_contigs, Scaffolds, small_scaffolds, Information, param):
    #search for linear streches in the graph, remove all cliques >2 and all contigs having more than three neighbors
    nr_edges = 0
    for edge in G.edges_iter():
        if G[edge[0]][edge[1]]['nr_links']:
            nr_edges += 1

    print >> Information, str(nr_edges) + ' link edges created.'
    print >> Information, 'Perform inference on scaffold graph...'

    #VizualizeGraph(G,param,Information)

    if param.detect_haplotype:
        HR.HaplotypicRegions(G, G_prime, Contigs, Scaffolds, param, Information)
    #save graph in dot format to file here

    ##If sigma specified. Pre calculate a look up table for every possible gap estimate in the common case
    ##where we have two long contigs that are linked. We do this one time per library to save computation time.
    if param.std_dev_ins_size:
        dValuesTable = GC.PreCalcMLvaluesOfdLongContigs(param.mean_ins_size, param.std_dev_ins_size, param.read_len)
    else :
        dValuesTable = None
    already_visited = set()
    if param.extend_paths:
        for node in G:
            already_visited.add(node)


    ##### Here is the scaffolding algorithm #######

    G = RemoveIsolatedContigs(G, Information)     #step1
    plot = 'G'

    if not param.no_score:
        RemoveAmbiguousRegionsUsingScore(G, G_prime, Information, param, plot) #step2
        G = RemoveIsolatedContigs(G, Information) #there are probably new isolated nodes created from step 2
        G, Contigs, Scaffolds = RemoveLoops(G, G_prime, Scaffolds, Contigs, Information, param)    #step4    
        #The contigs that made it to proper scaffolds
        (Contigs, Scaffolds, param) = NewContigsScaffolds(G, G_prime, Contigs, small_contigs, Scaffolds, small_scaffolds, Information, dValuesTable, param, already_visited)   #step5
        ##Here PathExtension algorithm between created scaffolds is called if PRO is activated

    if param.extend_paths:
        print >> Information, '\n\n\n Searching for paths BETWEEN scaffolds\n\n\n'
        PROBetweenScaf(G_prime, Contigs, small_contigs, Scaffolds, small_scaffolds, param, dValuesTable, Information)
        print >> Information, 'Nr of contigs left: ', len(G_prime.nodes()) / 2.0, 'Nr of linking edges left:', len(G_prime.edges()) - len(G_prime.nodes()) / 2.0
        print >> Information, 'Number of gaps estimated by GapEst-LP module pathgaps in this step is: {0}'.format(param.path_gaps_estimated)

    if param.plots:
        if len(param.gap_estimations) > 0:
            plots.histogram(param.gap_estimations, param, bins=50, x_label='gap size', y_label='Frequency', title='GapPredictions', nr_obs = len(param.gap_estimations))

 ## TODO: Look if we can use the new pathgaps theory to include these isolated 
 ## regions of small contigs in the scaffolding

#
#        for node in G_prime.nodes():
#            nbrs = G_prime.neighbors(node)
#            for nbr in nbrs:
#                if G_prime[node][nbr]['nr_links'] and 'score' not in G_prime[node][nbr]:
#                    G_prime.remove_edge(node, nbr)
#
#        plot = 'G_prime'
#        #print 'GOING IN:', len(G_prime.edges()), len(G_prime.nodes())
#        RemoveAmbiguousRegionsUsingScore(G_prime, G_prime, Information, param, plot)
#        G_prime, Contigs, Scaffolds = RemoveLoops(G_prime, G_prime, Scaffolds, Contigs, Information, param)
#        for node in G_prime:
#            if node[0] not in Scaffolds:
#                scaf_obj = small_scaffolds[node[0]]
#                Scaffolds[node[0]] = scaf_obj
#                cont_objects = scaf_obj.contigs
#                for obj_ in cont_objects:
#                    ctg_name = obj_.name
#                    Contigs[ctg_name] = obj_
#                    del small_contigs[ctg_name]
#                del small_scaffolds[node[0]]
#        (Contigs, Scaffolds, param) = NewContigsScaffolds(G_prime, G_prime, Contigs, small_contigs, Scaffolds, small_scaffolds, Information, dValuesTable, param, already_visited)


    ####### End of algorithm #####################


    return()



def RemoveIsolatedContigs(G, Information):
    print >> Information, 'Remove isolated nodes.'
    counter = 0
    for node in G.nodes():
        if node in G:
            nbr = G.neighbors(node)[0]
            if len(G.neighbors(node)) == 1 and len(G.neighbors(nbr)) == 1:
                counter += 1
                G.remove_nodes_from([node, nbr])
    print >> Information, str(counter) + ' isolated contigs removed from graph.'
    return(G)


def partition(pred, iterable):
    trues = []
    falses = []
    for item in iterable:
        if pred(item):
            trues.append(item)
        else:
            falses.append(item)
    return trues, falses

def remove_edges(G, G_prime, Information, param, node, score_chosen_obs, non_zero_score_removed_obs, edge_score_to_zero, record_decision):
    nbrs = G.neighbors(node)


    #Remove ambiguous edges
    filter_nbrs = filter(lambda nbr: 0 < G[node][nbr]['nr_links'] , nbrs) # remove other side of contig (has no "links")
    score_list_temp = sorted(map(lambda nbr: (G[node][nbr]['score'], nbr), filter_nbrs)) # sort list of scores for neighbors
    non_zero_score_edges, zero_score_edges = partition(lambda nbr: 0 < G[node][nbr[1]]['score'], score_list_temp) # Split list into 0-scoring and non-0-scoring edges


    #remove all zero score edges
    remove_zero_score_edges = map(lambda item: (node, item[1]) , zero_score_edges)
    G.remove_edges_from(remove_zero_score_edges)
    if param.plots:
        for item in zero_score_edges:
            edge_score_to_zero.append(item[0])

    if param.extend_paths:
        try: #we might have been removed this edge from G_prime when we did individual filtering of G_prime in CreateGraph module
            G_prime.remove_edges_from(remove_zero_score_edges)
        except nx.exception.NetworkXError:
            pass

    # Remove lower non zero score edges
    if len(non_zero_score_edges) > 1:
        if non_zero_score_edges[-2][0] / non_zero_score_edges[-1][0] > 0.9:
            print >> Information, 'SCORES AMBVIVALENT', non_zero_score_edges[-1][0], non_zero_score_edges[-2][0]
            remove_non_zero_edges = map(lambda item: (node, item[1]) , non_zero_score_edges) # Edges that does not have a score of 0 but are not the highest scoring one in an ambigous region
        else:
            remove_non_zero_edges = map(lambda item: (node, item[1]) , non_zero_score_edges[:-1]) # Edges that does not have a score of 0 but are not the highest scoring one in an ambigous region

        G.remove_edges_from(remove_non_zero_edges) # remove the lower scoring edges
        if param.extend_paths:
            try: #we might have been removed this edge from G_prime when we did individual filtering of G_prime in CreateGraph module
                G_prime.remove_edges_from(remove_non_zero_edges)
            except nx.exception.NetworkXError:
                pass

    if param.plots:
        if len(non_zero_score_edges) > 0:
            if record_decision:
                score_chosen_obs.append(non_zero_score_edges[-1][0])
            for item in non_zero_score_edges[:-1]:
                non_zero_score_removed_obs.append(item[0])


    return()

def RemoveAmbiguousRegionsUsingScore(G, G_prime, Information, param, plot):
    score_chosen_obs = []
    non_zero_score_removed_obs = []
    edge_score_to_zero = []

    nr_edges_before = len(G.edges())
    print >> Information, 'Remove edges from node if more than two edges'
    #counter1 = 0

    link_edges = filter(lambda x: 'score' in x[2], G.edges(data=True))
    edge_scores_sorted = sorted(link_edges, key=lambda x: x[2]['score'], reverse=True)

    for edge in edge_scores_sorted:
        remove_edges(G, G_prime, Information, param, edge[0], score_chosen_obs, non_zero_score_removed_obs, edge_score_to_zero, True)
        remove_edges(G, G_prime, Information, param, edge[1], score_chosen_obs, non_zero_score_removed_obs, edge_score_to_zero, False)

    nr_edges_after = len(G.edges())
    #print >> Information, str(counter1) + ' ambiguous regions in graph ( a contig with more than 2 neighbors).'
    print >> Information, ' Number of edges in G before:', nr_edges_before
    print >> Information, ' Number of edges in G after:', nr_edges_after
    try:
        print >> Information, ' %-age removed edges:', 100 * (1 - (nr_edges_after / float(nr_edges_before)))
    except ZeroDivisionError:
        pass

    if param.plots:
        print len(score_chosen_obs)
        list_of_datasets = [edge_score_to_zero, score_chosen_obs , non_zero_score_removed_obs]
        plots.multiple_histogram(list_of_datasets, param, 'score', 'frequency', title='Besst_decision_scores' + plot + '.' + param.bamfile.split('/')[-1])

    return()






def RemoveLoops(G, G_prime, Scaffolds, Contigs, Information, param):
#### After the proceure above, we hope that the graph is almost perfectly linear but we can still be encountering cycles (because of repeats or haplotypic contigs that has slipped through our conditions). Thus we finally search for loops
    print >> Information, 'Contigs/scaffolds left:', len(G.nodes()) / 2
    print >> Information, 'Remove remaining cycles...'
    graphs = nx.connected_component_subgraphs(G)
    #print 'Nr connected components',len(graphs)
    counter = 0
    for graph in graphs:
        list_of_cycles = algorithms.cycles.cycle_basis(graph)
        for cycle in list_of_cycles:
            print >> Information, 'A cycle in the scaffold graph: ' + str(cycle) + '\n'
            print >> Information, 'A cycle in the scaffold graph: ' + str(cycle), graph.edges()
            counter += 1
            for node in cycle:
                if node in G:
                    #we split up the whole cycle into separate contigs and send them to F
                    scaffold_ = node[0]
                    G.remove_nodes_from([(scaffold_, 'L'), (scaffold_, 'R')])
                    if param.extend_paths:
                        G_prime.remove_nodes_from([(scaffold_, 'L'), (scaffold_, 'R')])
#                    S_obj=Scaffolds[scaffold_]
#                    list_of_contigs=S_obj.contigs   #list of contig objects contained in scaffold object
#                    Contigs, F = GO.WriteToF(F,Contigs,list_of_contigs)
#                    del Scaffolds[scaffold_]
    print >> Information, str(counter) + ' cycles removed from graph.'
    return(G, Contigs, Scaffolds)

def NewContigsScaffolds(G, G_prime, Contigs, small_contigs, Scaffolds, small_scaffolds, Information, dValuesTable, param, already_visited):
### Remaining scaffolds are true sensible scaffolds, we must now update both the library of scaffold objects and the library of contig objects
    new_scaffolds_ = nx.connected_component_subgraphs(G)
    print >> Information, 'Nr of new scaffolds created in this step: ' + str(len(new_scaffolds_))
    for new_scaffold_ in new_scaffolds_:
        param.scaffold_indexer += 1
        #scaf_size=len(new_scaffold_)
        scaffold_length = 0
        contig_list = []

        ##### Here PathExtension algorithm is called if PRO is activated #####
        if param.extend_paths:
            PROWithinScaf(G, G_prime, Contigs, small_contigs, Scaffolds, small_scaffolds, param, new_scaffold_, dValuesTable, already_visited)

        for node in new_scaffold_:
            if len(G.neighbors(node)) == 1:
                start = node
                break
        for node in new_scaffold_:
            if len(G.neighbors(node)) == 1 and node != start:
                end = node
        #Create info to new scaffold object such as total length and the contig objects included

        prev_node = ('', '')
        pos = 0
        (G, contig_list, scaffold_length) = UpdateInfo(G, Contigs, small_contigs, Scaffolds, small_scaffolds, start, prev_node, pos, contig_list, scaffold_length, dValuesTable, param)
        S = Scaffold.scaffold(param.scaffold_indexer, contig_list, scaffold_length, defaultdict(constant_large), defaultdict(constant_large), defaultdict(constant_small), defaultdict(constant_small))  #Create the new scaffold object 

        Scaffolds[S.name] = S        #include in scaffold library

        if param.extend_paths:
            # Find the ends of the old subgraph new_scaffold_. We want them to be able to relabel these end nodes as the new sides on the new scaffold object created
            #only these ends are allowed to have links because they are of size mean+ 4*sigma so nothing is supposed to span over.

            #add the new scaffold object to G_prime

            G_prime.add_node((S.name, 'L'))  #start node
            G_prime.add_node((S.name, 'R'))  # end node
            G_prime.add_edge((S.name, 'L'), (S.name, 'R'), nr_links=None)
            try:
                for nbr in G_prime.neighbors(start):
                    nr_links_ = G_prime[start][nbr]['nr_links']
                    if nr_links_:
                        obs_ = G_prime[start][nbr]['obs']
                        G_prime.add_edge((S.name, 'L'), nbr, nr_links=nr_links_, obs=obs_)

                for nbr in G_prime.neighbors(end):
                    nr_links_ = G_prime[end][nbr]['nr_links']
                    if nr_links_:
                        obs_ = G_prime[end][nbr]['obs']
                        G_prime.add_edge((S.name, 'R'), nbr, nr_links=nr_links_, obs=obs_)

                #remove the old scaffold objects from G_prime
                G_prime.remove_nodes_from(new_scaffold_)
            except nx.exception.NetworkXError:
                pass

    return(Contigs, Scaffolds, param)


def UpdateInfo(G, Contigs, small_contigs, Scaffolds, small_scaffolds, node, prev_node, pos, contig_list, scaffold_length, dValuesTable, param):
    scaf = node[0]
    side = node[1]
    prev_scaf = prev_node[0]
    if len(G.neighbors((scaf, side))) == 0:  #reached end of scaffol
        #find the contig with the largest position
        object_with_largest_pos_in_scaffold = max(contig_list, key=lambda object: object.position + object.length)
        scaffold_length = object_with_largest_pos_in_scaffold.position + object_with_largest_pos_in_scaffold.length
#        try:
        del Scaffolds[scaf] #finally, delete the old scaffold object
        G.remove_node((scaf, side))
#        except KeyError:
#            del small_scaffolds[scaf] #finally, delete the old scaffold object
        return(G, contig_list, scaffold_length)
    else:
        nbr_node = G.neighbors((scaf, side))
        nbr_scaf = nbr_node[0][0]
        nbr_side = nbr_node[0][1]
        if scaf != prev_scaf:
            if side == 'L':    #Contig/scaffold still has same orientation as in previous iteration, just update position in scaffold                                           
                #want to assign nr of links to contig object, note that in case of a "multiple contigs"-scaffold object, only the outermost contig holds the information of the total nr of links between the two scaffold objects
                #try:
                contig_objects = Scaffolds[scaf].contigs #list of contig objects
                #except KeyError:
                #    contig_objects=small_scaffolds[scaf].contigs #list of contig objects
                #Update just update position in scaffold 
                for contig in contig_objects:
                    contig.scaffold = param.scaffold_indexer
                    contig.position += pos
                    #direction unchanged
                    contig_list.append(contig)
                G.remove_node((scaf, side))
                prev_node = node
                node = (nbr_scaf, nbr_side)
                #try:                 
                pos += Scaffolds[scaf].s_length  #update position before sending it to next scaffold
                #except KeyError:
                #    pos+=small_scaffolds[scaf].s_length  #update position before sending it to next scaffold

                G, contig_list, scaffold_length = UpdateInfo(G, Contigs, small_contigs, Scaffolds, small_scaffolds, node, prev_node, pos, contig_list, scaffold_length, dValuesTable, param)

            else:  #Contig/scaffold need to change orientation as well as modify orientation in this case
                #try:
                contig_objects = Scaffolds[scaf].contigs #list of contig objects
                #except KeyError:
                #    contig_objects=small_scaffolds[scaf].contigs #list of contig objects

                for contig in contig_objects:
                    contig.scaffold = param.scaffold_indexer
                    #try:
                    curr_scaf_length = Scaffolds[scaf].s_length
                    #except KeyError:
                    #    curr_scaf_length=small_scaffolds[scaf].s_length

                    curr_pos_within_scaf = contig.position
                    contig_length = contig.length
                    contig.position = pos + (curr_scaf_length - curr_pos_within_scaf) - contig_length #updates the position within scaf
                    contig.direction = bool(True -contig.direction) #changes the direction
                    contig_list.append(contig)

                G.remove_node((scaf, side))
                prev_node = node
                node = (nbr_scaf, nbr_side)
                #try:                    
                pos += Scaffolds[scaf].s_length  #update position before sending it to next scaffold
                #except KeyError:
                #pos+=small_scaffolds[scaf].s_length  #update position before sending it to next scaffold

                G, contig_list, scaffold_length = UpdateInfo(G, Contigs, small_contigs, Scaffolds, small_scaffolds, node, prev_node, pos, contig_list, scaffold_length, dValuesTable, param)
        else:
            if  'avg_gap' not in  G[(scaf, side)][(nbr_scaf, nbr_side)]:
                #calculate gap to next scaffold
                sum_obs = G[(scaf, side)][(nbr_scaf, nbr_side)]['obs']
                nr_links = G[(scaf, side)][(nbr_scaf, nbr_side)]['nr_links']
                data_observation = (nr_links * param.mean_ins_size - sum_obs) / float(nr_links)
                mean_obs = sum_obs / float(nr_links)
                #try:
                c1_len = Scaffolds[scaf].s_length
                #except KeyError:
                #    c1_len=small_scaffolds[scaf].s_length
                #try:
                c2_len = Scaffolds[nbr_scaf].s_length
                #except KeyError:
                #   c2_len=small_scaffolds[nbr_scaf].s_length
                #do fancy gap estimation by the bias estimator formula
                if param.std_dev_ins_size and nr_links >= 5:
                    #pre calculated value in lookup table 
                    if c1_len > param.mean_ins_size + 4 * param.std_dev_ins_size and c2_len > param.mean_ins_size + 4 * param.std_dev_ins_size:
                        #(heuristic scale down of table to gaps of at most 2 stddevs away from mean)
                        try:
                            avg_gap = dValuesTable[int(round(data_observation, 0))]
                        except KeyError:
                            avg_gap = GC.GapEstimator(param.mean_ins_size, param.std_dev_ins_size, param.read_len, mean_obs, c1_len, c2_len)
                            #print 'Gap estimate was outside the boundary of the precalculated table, obs were: ', data_observation, 'binary search gave: ', avg_gap
                    #Do binary search for ML estimate of gap if contigs is larger than 3 times the std_dev
                    elif c1_len > param.std_dev_ins_size + param.read_len and c2_len > param.std_dev_ins_size + param.read_len:
                        avg_gap = GC.GapEstimator(param.mean_ins_size, param.std_dev_ins_size, param.read_len, mean_obs, c1_len, c2_len)
                    else:
                        #print 'now', 2 * param.std_dev_ins_size + param.read_len
                        avg_gap = int(data_observation)
                        param.gap_estimations.append( avg_gap )
                    #print 'Gapest if used:' + str(int(avg_gap)), 'Naive: ' + str(int(data_observation)), c1_len, c2_len, Scaffolds[scaf].contigs[0].name, Scaffolds[nbr_scaf].contigs[0].name
                    #See if the two contigs are in fact negatively overlapped in the delta file, , then abyss produses
                    #contigs contained in other contigs
                #do naive gap estimation
                else:
                    avg_gap = int(data_observation)
                    param.gap_estimations.append( avg_gap )
            else:
                avg_gap = G[(scaf, side)][(nbr_scaf, nbr_side)]['avg_gap']
                param.gap_estimations.append( avg_gap )

            if avg_gap <= 1:
                #TODO: Eventually implement SW algm to find ML overlap
                avg_gap = 1

            pos += int(avg_gap)
            G.remove_node((scaf, side))
            prev_node = node
            node = (nbr_scaf, nbr_side)
            #try:
            del Scaffolds[scaf] #finally, delete the old scaffold object
            #except KeyError:
            #    del small_scaffolds[scaf] #finally, delete the old scaffold object
            G, contig_list, scaffold_length = UpdateInfo(G, Contigs, small_contigs, Scaffolds, small_scaffolds, node, prev_node, pos, contig_list, scaffold_length, dValuesTable, param)
    return(G, contig_list, scaffold_length)

def PROWithinScaf(G, G_prime, Contigs, small_contigs, Scaffolds, small_scaffolds, param, new_scaffold_, dValuesTable, already_visited):
    #loc_count = 0
    for edge in new_scaffold_.edges_iter():
        nr_links_ = G[edge[0]][edge[1]]['nr_links']
        if nr_links_:
            start = edge[0]
            end = edge[1]
            c1_len = Scaffolds[start[0]].s_length
            c2_len = Scaffolds[end[0]].s_length
            sum_obs = G[start][end]['obs']
            data_observation = (nr_links_ * param.mean_ins_size - sum_obs) / float(nr_links_)
            avg_gap = GC.GapEstimator(param.mean_ins_size, param.std_dev_ins_size, param.read_len, data_observation, c1_len, c2_len)

            #high_score_path, bad_links, score, path_len = ELS.WithinScaffolds(G, G_prime, start, end, already_visited, param.ins_size_threshold, param)
            all_paths_sorted_wrt_score = ELS.WithinScaffolds(G, G_prime, start, end, already_visited, param.ins_size_threshold, param)

            if not all_paths_sorted_wrt_score:
                continue

            # if all_paths_sorted_wrt_score[-1][1] != 0:
            #     sublist = all_paths_sorted_wrt_score[-1]
            #     bad_links = sublist[1]
            #     score = sublist[0]
            #     path_len = sublist[3]
            #     print 'Path: path length: {0}, nr bad links: {1}, score: {2} '.format((path_len - 2) / 2.0, bad_links, score)

            # # for sublist in reversed(all_paths_sorted_wrt_score):
            # #     path = sublist[2]
            # #     bad_links = sublist[1]
            # #     score = sublist[0]
            # #     path_len = sublist[3]
            # #     print 'Path: path length: {0}, nr bad links: {1}, score: {2} '.format((path_len - 2) / 2.0, bad_links, score)

            high_score_path, score = all_paths_sorted_wrt_score[-1][2], all_paths_sorted_wrt_score[-1][0]
            #print 'Highest scoring path:{0}'.format(high_score_path)
            #print 'Score: {0}'.format(score)

            if high_score_path and score >= 0.0:

                ##################### v1.0.4.5 
                ## modified improved path gap estimation here!!

                high_score_path_copy = copy.deepcopy(high_score_path)
                G_, path = estimate_path_gaps(high_score_path_copy,Scaffolds,small_scaffolds, G_prime,param)
                del path[0]
                del path[-1]
                G.remove_edge(start, end)
                G.add_edges_from(G_.edges(data=True))
                G_prime.remove_nodes_from(path[1:-1])

                # also remove the edges from inner ends of the large scaffolds so they can't be
                # involved in creating paths in "between scaffolds"

                G_prime.remove_node(path[0])
                G_prime.remove_node(path[-1])
                if path[0][1] == 'L':
                    G_prime.add_edge(path[0],(path[0][0],'R'),nr_links=None)
                else:
                    G_prime.add_edge(path[0],(path[0][0],'L'),nr_links=None)
                if path[-1][1] == 'L':
                    G_prime.add_edge(path[-1],(path[-1][0],'R'),nr_links=None)
                else:
                    G_prime.add_edge(path[-1],(path[-1][0],'L'),nr_links=None)

                # move all contig and scaffold objects from "small" structure to large structure to fit with UpdateInfo structure
                small_scafs = map(lambda i: path[i], filter(lambda i: i % 2 == 1, range(len(path) - 1)))
                for item in small_scafs:
                    scaf_obj = small_scaffolds[item[0]]
                    Scaffolds[item[0]] = scaf_obj
                    cont_objects = scaf_obj.contigs
                    for obj_ in cont_objects:
                        ctg_name = obj_.name
                        Contigs[ctg_name] = obj_
                        del small_contigs[ctg_name]
                    del small_scaffolds[item[0]]
                #sys.exit()
                ######################################



                # #loc_count += 1
                # #remove edge in G to fill in the small scaffolds
                # G.remove_edge(start, end)
                # #add small scaffolds to G
                # for i in range(0, len(high_score_path) - 1):
                #     nr_lin = G_prime[high_score_path[i]][high_score_path[i + 1]]['nr_links']
                #     try:
                #         total_dist = G_prime[high_score_path[i]][high_score_path[i + 1]]['obs']
                #         G.add_edge(high_score_path[i], high_score_path[i + 1], nr_links=nr_lin, obs=total_dist)
                #     except KeyError:
                #         G.add_edge(high_score_path[i], high_score_path[i + 1], nr_links=nr_lin)
                # #remove the small contigs from G_prime
                # G_prime.remove_nodes_from(high_score_path[1:-1])
                # # move all contig and scaffold objects from "small" structure to large structure to fit with UpdateInfo structure
                # small_scafs = map(lambda i: high_score_path[i], filter(lambda i: i % 2 == 1, range(len(high_score_path) - 1)))
                # for item in small_scafs:
                #     scaf_obj = small_scaffolds[item[0]]
                #     Scaffolds[item[0]] = scaf_obj
                #     cont_objects = scaf_obj.contigs
                #     for obj_ in cont_objects:
                #         ctg_name = obj_.name
                #         Contigs[ctg_name] = obj_
                #         del small_contigs[ctg_name]
                #     del small_scaffolds[item[0]]

    ####################################################################
    return()


def other_end(node):
    if node[1] == 'L':
        return (node[0],'R')
    elif node[1] == 'R':
        return (node[0],'L')
    else:
        print 'Node is not properly declared (has not right or left end)'
        sys.exit()

def permute_path(path, ctg_to_move, contig_after):
    pos_ctg_to_move_left = path.index((ctg_to_move,'L'))
    pos_ctg_to_move_right = path.index((ctg_to_move,'R'))

    path = filter(lambda x: x[0] != ctg_to_move, path)
    pos_of_ctg_after = map(lambda x: x[0],path).index(contig_after)
    if pos_ctg_to_move_right < pos_ctg_to_move_left:
        path.insert(pos_of_ctg_after, (ctg_to_move,'L') )
        path.insert(pos_of_ctg_after, (ctg_to_move,'R') )
    else:
        path.insert(pos_of_ctg_after, (ctg_to_move,'R') )
        path.insert(pos_of_ctg_after, (ctg_to_move,'L') )

    return path


def calculate_path_LP(current_path,Scaffolds,small_scaffolds,observations,param,initial_path):
    contigs_to_indexes = {}
    indexes_to_contigs = {}
    index = 0
    #print 'CURNNNT PATH_',current_path
    for ctg in current_path:
        if ctg[0] in contigs_to_indexes:
            continue
        else:
            contigs_to_indexes[ctg[0]] = index
            indexes_to_contigs[index] =  ctg[0]

            index += 1

    contig_lengths = []
    for ctg in contigs_to_indexes:
        try:
            contig_lengths.append((contigs_to_indexes[ctg], Scaffolds[ctg].s_length))
        except KeyError:
            contig_lengths.append((contigs_to_indexes[ctg], small_scaffolds[ctg].s_length))

    #contig_lengths2 = map(lambda x: (contigs_to_indexes[x], Scaffolds[x].s_length), contigs_to_indexes) 

    #print contig_lengths
    ctg_lengths_sorted = map(lambda x: x[1], sorted(contig_lengths, key=lambda x: x[0]))
    #print ctg_lengths_sorted
    tot_links = 0
    mp_links = 0
    index_observations = {}     
    for c1,c2 in observations:  
        tot_links += len(observations[(c1,c2)])  
        if current_path.index(c1) < current_path.index(c2) and current_path.index(c1) % 2 == 1 and current_path.index(c2) % 2 == 0 and param.contamination_ratio:
            PE = 1
            #print 'PE link!!',c1,c2
        elif current_path.index(c1) > current_path.index(c2) and current_path.index(c1) % 2 == 0 and current_path.index(c2) % 2 == 1 and param.contamination_ratio:
            PE = 1
            #print 'PE link!!',c1,c2
        elif current_path.index(c1) < current_path.index(c2) and current_path.index(c1) % 2 == 0 and current_path.index(c2) % 2 == 1:
            PE = 0
            mp_links += len(observations[(c1,c2)])
            #print 'MP link!!',c1,c2
        elif current_path.index(c1) > current_path.index(c2) and current_path.index(c1) % 2 == 1 and current_path.index(c2) % 2 == 0:
            mp_links += len(observations[(c1,c2)])
            PE = 0
            #print 'MP link!!',c1,c2           
        else:
            PE = -1
            #print 'Spurious link!!',c1,c2

        if PE == -1:
            # we dont want to have a spurious link to contiribute to the LP objective value
            continue
        else:
            i1, i2 = min(contigs_to_indexes[c1[0]], contigs_to_indexes[c2[0]]), max(contigs_to_indexes[c1[0]], contigs_to_indexes[c2[0]])
            index_observations[(i1,i2,PE)] = observations[(c1,c2)]
        #print i1,i2 #'OBSLIST_', observations[(c1,c2)]
    #observations =  map(lambda x: (contigs_to_indexes[x[0][0]], contigs_to_indexes[x[1][0]]) = observations[x], observations)   

    #print 'MP LINK RATIO:{0}, tot_links:{1}'.format(mp_links/float(tot_links), tot_links)
    #print index_observations
    #print ctg_lengths_sorted
    #print index_observations

    ## 2 Get optimal LP solution for given path order
    result_path = order_contigs.main(ctg_lengths_sorted, index_observations, param, initial_path)
    if not result_path:
        return None,None,None,None

    return result_path, contigs_to_indexes, indexes_to_contigs, index_observations


def estimate_path_gaps(path,Scaffolds,small_scaffolds, G_prime, param):
    ## ACCURATE GAP EST HERE

    #print G_prime.subgraph(path).edges(data=True)
    sub_graph = G_prime.subgraph(path)


    #print G_prime.subgraph(path).edges(data=True)
    sub_graph_reduced = filter(lambda x: sub_graph[x[0]][x[1]]['nr_links'] != None and x[0][0] in sub_graph[x[0]][x[1]] , sub_graph.edges())

    observations = dict(map(lambda x: (x, [i+j for i,j in zip(sub_graph[x[0]][x[1]][x[0][0]], sub_graph[x[0]][x[1]][x[1][0]] )]), sub_graph_reduced))
    sub_graph_small_to_large_ctgs = filter(lambda x: sub_graph[x[0]][x[1]]['nr_links'] != None and x[0][0] not in sub_graph[x[0]][x[1]] , sub_graph.edges())
    for c1,c2 in sub_graph_small_to_large_ctgs:
        observations[(c1,c2)] = [sub_graph[c1][c2]['obs']/ sub_graph[c1][c2]['nr_links']]*sub_graph[c1][c2]['nr_links']
    

    # for c1,c2 in observations:
    #     if (other_end(c2),other_end(c1)) in observations:
    #         print observations
    # print param.contamination_ratio 



    #########################
    ##########################
    ## Now we have all info that we need to send to module for calculating optimal path



    # only one contig, nothing to permute
    if len(path) <= 4 or not param.contamination_ratio:
        final_path_instance, final_contigs_to_indexes, final_indexes_to_contigs, final_index_observations = calculate_path_LP(path,Scaffolds,small_scaffolds,observations,param,True)
        final_path = path
        #print final_path
        
    ## algm here
    else:
        #print 'ENTER HERE'
        #print path
        final_path_instance, final_contigs_to_indexes, final_indexes_to_contigs, final_index_observations = calculate_path_LP(path,Scaffolds,small_scaffolds,observations,param, True)
        final_path = copy.deepcopy(path)
        original_path = copy.deepcopy(path)
        #print 'WORK IS DONE'

        for i in range(3, len(path) - 1, 2):
            # switch positions of two contigs
            current_path = copy.deepcopy(final_path)
            ctg_to_move = original_path[i][0]
            contig_after = original_path[i-2][0]
            current_path  = permute_path(current_path, ctg_to_move, contig_after)

            #print 'Current path:',current_path
        ## 1 Get a mapping from contigs to indexes (index for contig order in the current path)
            current_path_instance, current_contigs_to_indexes, current_indexes_to_contigs, current_index_observations = calculate_path_LP(current_path,Scaffolds,small_scaffolds,observations,param,False)
            if not current_path_instance:
                continue
        ## 3 Check of current path is better than previous
            #print 'Current objective: {0}, best objective: {1}'.format(current_path_instance.objective, final_path_instance.objective)
            
            if current_path_instance.objective < final_path_instance.objective:
                final_path = copy.deepcopy(current_path)
                final_path_instance = copy.deepcopy(current_path_instance)
                final_contigs_to_indexes = current_contigs_to_indexes 
                final_indexes_to_contigs = current_indexes_to_contigs
                final_index_observations = current_index_observations


        # for i in range(1,len(path) - 3, 2):
        #     #final_path = copy.deepcopy(path)
        #     for j in range(i+2, len(path) - 1, 2):
        #         # switch positions of two contigs
        #         current_path = copy.deepcopy(final_path)
        #         ctg_to_move = final_path[j][0]
        #         current_pos_examined_contig = final_path[i][0]
        #         current_path  = permute_path(current_path, ctg_to_move, current_pos_examined_contig)
        #         print 'Current path:',current_path
        #     ## 1 Get a mapping from contigs to indexes (index for contig order in the current path)
        #         current_path_instance, current_contigs_to_indexes, current_indexes_to_contigs, current_index_observations = calculate_path_LP(current_path,Scaffolds,small_scaffolds,observations,param)

        #     ## 3 Check of current path is better than previous
        #         print 'Current objective: {0}, best objective: {1}'.format(current_path_instance.objective, final_path_instance.objective)
        #         if current_path_instance.objective < final_path_instance.objective:
        #             final_path = copy.deepcopy(current_path)
        #             final_path_instance = copy.deepcopy(current_path_instance)
        #             final_contigs_to_indexes = current_contigs_to_indexes 
        #             final_indexes_to_contigs = current_indexes_to_contigs
        #             final_index_observations = current_index_observations




    ## 5 Calculate some stats on the path that we have chosen

    param.path_gaps_estimated += len(final_path_instance.gaps)
    path_dict_index = final_path_instance.make_path_dict_for_besst()
    #print 'FINAL:', final_path
    #print 'OBJECTIVE: {0}'.format(final_path_instance.objective)

    # for ctg in final_path:
    #     if ctg[0] in Scaffolds:
    #         for c in Scaffolds[ctg[0]].contigs:
    #             print 'ctg name, scaffold{0}: {1}'.format(ctg[0],c.name)
    #     else:
    #         print 'small ctg name:', small_scaffolds[ctg[0]].contigs[0].name


    #print final_path_instance
    #print path_dict_index
    path_dict = map(lambda x: (final_indexes_to_contigs[x[0].index],final_indexes_to_contigs[x[1].index], path_dict_index[x]), path_dict_index)
    #print path_dict
    # for ctg in final_path_instance.ctgs:
    #     contig = indexes_to_contigs[ctg.index]
    #     ctg.length
    #     ctg.position

    G_ = nx.Graph()
    final_path.insert(0, (final_path[0][0], 'R')) if final_path[0][1] == 'L' else final_path.insert(0, (final_path[0][0], 'L'))
    final_path.insert(len(final_path), (final_path[-1][0], 'R'))  if final_path[-1][1] == 'L' else final_path.insert(len(final_path), (final_path[-1][0], 'L'))
    G_.add_edges_from(zip(final_path[::1], final_path[1::]))

    for c1,c2,gap in path_dict:
        if (c2,'L') in G_[(c1,'L')]:
            G_[(c1,'L')][(c2,'L')]['avg_gap'] = gap 
        elif (c2,'R') in G_[(c1,'L')]:
            G_[(c1,'L')][(c2,'R')]['avg_gap'] = gap
        elif (c2,'L') in G_[(c1,'R')]:
            G_[(c1,'R')][(c2,'L')]['avg_gap'] = gap
        elif (c2,'R') in G_[(c1,'R')]:
            G_[(c1,'R')][(c2,'R')]['avg_gap'] = gap
        else:
            print 'Could not find edge!'
            sys.exit()
    #print G_.edges(data=True)

    return(G_, final_path)


def PROBetweenScaf(G_prime, Contigs, small_contigs, Scaffolds, small_scaffolds, param, dValuesTable, Information):
    start_scaf_index = param.scaffold_indexer
    G = nx.Graph()
    for node in G_prime:
        if node[0] in Scaffolds: # meets the length criteria
            G.add_node(node)

    # Filtering and heuristic here to reduce computation if needed O(n^2) in contigs on pathfinder

    #remove all solated contigs
    for node in G.nodes():
        if node in G:
            nbr = G_prime.neighbors(node)[0]
            if len(G_prime.neighbors(node)) == 1 and len(G_prime.neighbors(nbr)) == 1:
                G.remove_nodes_from([node, nbr])


    if len(G.nodes()) / 2.0 > 10000:
        # Too few short contigs compared to long (ratio set to 0.1) or lib ins size + 2*std_dev - 2*read_len < 200 ) and too many large contigs (> 10 000) do not enter path extension algm since to low payoff:

        if len(small_scaffolds) / float(len(Scaffolds)) < 0.1:
            print >> Information, "Did not enter path seartching algorithm between scaffolds due to too small fraction of small scaffolds, fraction were: ", len(small_scaffolds) / float(len(Scaffolds))
            return(start_scaf_index)

    ########### Find paths between scaffolds here ###############

    # Multi Processing (if available), check nr of available cores
    num_cores = multiprocessing.cpu_count()
    #TODO: If we get too many paths back and run into memory issues we could change so that only paths with score over 0 are stored in ELS module
    if param.multiprocess and num_cores > 1:
        import workerprocess
        import heapq
        print >> Information, 'Entering ELS.BetweenScaffolds parallelized with ', num_cores, ' cores.'
        start = time.time()
        # load up work queue
        work_queue = multiprocessing.Queue()
        end = set()
        for node in G:
            end.add(node)
        nodes = G.nodes()
        nr_jobs = len(nodes)
        chunk = nr_jobs / (num_cores)
        counter = 0
        nr_processes = 0
        # partition equally many nodes in G to each core
        while counter < nr_jobs:
            work_queue.put((set(nodes[counter:counter + chunk]), G_prime, end, param))
            nr_processes += 1
            print >> Information, 'node nr', counter, 'to', counter + chunk - 1, 'added'
            #print work_queue.get()
            counter += chunk

        # create a queue to pass to workers to store the results
        result_queue = multiprocessing.Queue()

        # spawn workers
        while not work_queue.empty():
            worker = workerprocess.Worker(work_queue.get(), result_queue)
            worker.start()

        # collect the results off the queue
        results = []
        for i in range(nr_processes):
            res = result_queue.get()
            results.append(res)

        def wrapper(func, args):
            return(func(*args))
        all_paths_sorted_wrt_score_itr = wrapper(heapq.merge, results) #tot_result
        all_paths_sorted_wrt_score = [i for i in all_paths_sorted_wrt_score_itr]
        elapsed = time.time() - start
        print >> Information, "Elapsed time multiprocessing: ", elapsed

    else:
        start = time.time()
        end = set()
        for node in G:
            end.add(node)
        iter_nodes = end.copy()
        print >> Information, 'Entering ELS.BetweenScaffolds single core'
        all_paths_sorted_wrt_score = ELS.BetweenScaffolds(G_prime, end, iter_nodes, param)
        elapsed = time.time() - start
        print >> Information, "Elapsed time single core pathfinder: ", elapsed

    ################################################################

    start_end_node_update_storage = {}
    print >> Information, 'Total number of paths between scaffolds detected:', len(all_paths_sorted_wrt_score)
    for sublist in reversed(all_paths_sorted_wrt_score):
        path = sublist[2]
        bad_links = sublist[1]
        score = sublist[0]
        path_len = sublist[3]
        print >> Information, 'Path: path length: {0}, nr bad links: {1}, score: {2} '.format((path_len - 2) / 2.0, bad_links, score)

        ## Need something here that keeps track on which contigs that are added to Scaffolds so that a
        ## contig is only present once in each path

        #print start_end_node_update_storage
        # Either a small contig/scaffold has been included in a path earlier and thus has moved it's object to Scaffolds (and changed index) 
        small_scaf_is_already_in = 0
        for scaf_ in path[1:-1]:
            if scaf_[0] not in small_scaffolds:
                small_scaf_is_already_in = 1
                #print 'At least one of the contigs is already in another scaffold'
                break
        if small_scaf_is_already_in:
            continue

        # A very special corner case (circular paths)
        if path[0][0] not in Scaffolds and path[-1][0] not in Scaffolds:
            try:
                strt = start_end_node_update_storage[path[0]][0]
                nd = start_end_node_update_storage[path[-1]][0]
                if strt[0] == nd[0]:
                    print >> Information, 'Rare case (circular paths) detected and treated. '
                    continue
            except KeyError:
                pass

        # Or a large scaffold/contig has changed scaffold index due to one of it's sides is present in another path (we still want to allow for paths from the other side)
        case1 = 0
        case2 = 0
        if path[0][0] not in Scaffolds:
            if path[0] in start_end_node_update_storage:
                case1 = 1
            else:
                print >> Information, 'Beginning is already in path'
                continue

        if path[-1][0] not in Scaffolds:
            if path[-1] in start_end_node_update_storage:
                case2 = 1
            else:
                print >> Information, 'End is already in path'
                continue


        original_start_node = path[0]

        if path[0][0] not in Scaffolds:
            #large scaffold has changed index before. This suggested path is however from it's other side
            node_to_remove1 = path[0]
            path[0] = start_end_node_update_storage[node_to_remove1][0]
            #update the node on the other end of the end scaffold to point at the newest index
            node_to_refresh1 = start_end_node_update_storage[node_to_remove1][1]
            #print 'Enter 1'
            try:
                node_ptr = start_end_node_update_storage[ path[-1] ][1]
                #print '1.1', node_ptr,start_end_node_update_storage[ path[-1] ]
            except KeyError:
                other_side = 'L' if path[-1][1] == 'R' else 'R'
                node_ptr = (path[-1][0], other_side)
                #print '1.2', node_ptr, path[-1]
            start_end_node_update_storage[node_to_refresh1] = [(param.scaffold_indexer + 1, 'L'), node_ptr  ]
            #path pointer can be accesed only once needs to be destroyed after
            del start_end_node_update_storage[node_to_remove1]


        if path[-1][0] not in Scaffolds:
            #large scaffold has changed index before. This suggested path is however from it's other side
            #print 'case2.2'
            node_to_remove2 = path[-1]
            path[-1] = start_end_node_update_storage[node_to_remove2][0]
            #update the node on the other end of the end scaffold to point at the newest index
            node_to_refresh2 = start_end_node_update_storage[node_to_remove2][1]
            #print 'Enter 2'
            try:
                node_ptr = start_end_node_update_storage[ original_start_node ][1]
                #print '2.1', node_ptr, start_end_node_update_storage[ original_start_node ]
            except KeyError:
                other_side = 'L' if original_start_node[1] == 'R' else 'R'
                node_ptr = (original_start_node[0], other_side)
                #print '2.2', node_ptr,original_start_node          
            start_end_node_update_storage[node_to_refresh2] = [(param.scaffold_indexer + 1, 'R'), node_ptr ]
            #path pointer can be accesed only once needs to be destroyed after
            del start_end_node_update_storage[node_to_remove2]


        # Here we update the contigs that lies in small_contigs to Contigs. We need to do this here because
        # we update the scaffold index below

        # move all contig and scaffold objects from "small" structure to large structure to fit with UpdateInfo structure

        small_scafs = map(lambda i: path[i], filter(lambda i: i % 2 == 1, range(len(path) - 1)))
        for item in small_scafs:
            scaf_obj = small_scaffolds[item[0]]
            Scaffolds[item[0]] = scaf_obj
            cont_objects = scaf_obj.contigs
            for obj_ in cont_objects:
                ctg_name = obj_.name
                Contigs[ctg_name] = obj_
                del small_contigs[ctg_name]
            del small_scaffolds[item[0]]
        ## Here we do the "joining of two scaffolds with the new path if no contig/scaffold is present
        ## in another path, we need to update "Scaffolds" structure here along as we go in order for
        ## the above dublette checking function to work


        G_, path = estimate_path_gaps(path,Scaffolds,small_scaffolds, G_prime,param)


        # #make the path a small linear graph
        # G_ = nx.Graph()
        # path.insert(0, (path[0][0], 'R')) if path[0][1] == 'L' else path.insert(0, (path[0][0], 'L'))
        # path.insert(len(path), (path[-1][0], 'R'))  if path[-1][1] == 'L' else path.insert(len(path), (path[-1][0], 'L'))
        # G_.add_edges_from(zip(path[::1], path[1::]))

        start_end_node_update_storage[path[0]] = 0
        start_end_node_update_storage[path[-1]] = 0

        # for edge in G_.edges():
        #     try:
        #         G_[edge[0]][edge[1]]['nr_links'] = G_prime[edge[0]][edge[1]]['nr_links']
        #     except KeyError:
        #         print >> Information, path
        #         try:
        #             Scaffolds[edge[0][0]]
        #             print >> Information, edge[0][0] , 'is in Scaffolds'
        #         except KeyError:
        #             print >> Information, edge[0][0] , 'is not in Scaffolds'
        #         try:
        #             Scaffolds[edge[1][0]]
        #             print >> Information, edge[1][0] , 'is in Scaffolds'
        #         except KeyError:
        #             print >> Information, edge[1][0] , 'is not in Scaffolds'

        #         try:
        #             small_scaffolds[edge[0][0]]
        #             print >> Information, edge[0][0] , 'is in small_scaffolds'
        #         except KeyError:
        #             print >> Information, edge[0][0] , 'is not in small_scaffolds'
        #         try:
        #             small_scaffolds[edge[1][0]]
        #             print >> Information, edge[1][0] , 'is in small_scaffolds'
        #         except KeyError:
        #             print >> Information, edge[1][0] , 'is not in small_scaffolds'

        #         try:
        #             G_prime[edge[0]]
        #             print >> Information, edge[0] , 'is in G_prime'
        #             print >> Information, G_prime[edge[0]]
        #         except KeyError:
        #             print >> Information, edge[0] , 'is not in G_prime'
        #         try:
        #             G_prime[edge[1]]
        #             print >> Information, edge[1] , 'is in G_prime'
        #             print >> Information, G_prime[edge[1]]
        #         except KeyError:
        #             print >> Information, edge[1] , 'is not in G_prime'
        #         print edge[0], edge[1]
        #         G_[edge[0]][edge[1]]['nr_links'] = G_prime[edge[0]][edge[1]]['nr_links']
        #         sys.exit()

        #     try:
        #         G_[edge[0]][edge[1]]['obs'] = G_prime[edge[0]][edge[1]]['obs']
        #     except KeyError:
        #         #may be the two different sides of a contig (has no gap dist)
        #         pass

        start = path[0]
        end = path[-1]
        prev_node = ('', '')
        pos = 0
        scaffold_length = 0
        contig_list = []
        param.scaffold_indexer += 1
        (G, contig_list, scaffold_length) = UpdateInfo(G_, Contigs, small_contigs, Scaffolds, small_scaffolds, start, prev_node, pos, contig_list, scaffold_length, dValuesTable, param)
        S = Scaffold.scaffold(param.scaffold_indexer, contig_list, scaffold_length, defaultdict(constant_large), defaultdict(constant_large), defaultdict(constant_small), defaultdict(constant_small))  #Create the new scaffold object 

        print >> Information, 'Path taken! path length: {0}, nr bad links: {1}, score: {2} '.format((path_len - 2) / 2.0, bad_links, score)

        Scaffolds[S.name] = S        #include in scaffold library
        #add the new scaffold object to G_prime

        G_prime.add_node((S.name, 'L'))  #start node
        G_prime.add_node((S.name, 'R'))  # end node
        G_prime.add_edge((S.name, 'L'), (S.name, 'R'), nr_links=None)
        for nbr in G_prime.neighbors(start):
            nr_links_ = G_prime[start][nbr]['nr_links']
            if nr_links_:
                obs_ = G_prime[start][nbr]['obs']
                G_prime.add_edge((S.name, 'L'), nbr, nr_links=nr_links_, obs=obs_)

        for nbr in G_prime.neighbors(end):
            nr_links_ = G_prime[end][nbr]['nr_links']
            if nr_links_:
                obs_ = G_prime[end][nbr]['obs']
                G_prime.add_edge((S.name, 'R'), nbr, nr_links=nr_links_, obs=obs_)

        #remove the old scaffold objects from G_prime
        G_prime.remove_nodes_from(path)

        #updated beginning
        if case1 and not case2:
            start_end_node_update_storage[node_to_refresh1] = [(S.name, 'L'), path[-1] ]
            start_end_node_update_storage[path[-1]] = [(S.name, 'R'), node_to_refresh1 ]
        elif case2 and not case1:
            start_end_node_update_storage[path[0]] = [(S.name, 'L'), node_to_refresh2 ]
            start_end_node_update_storage[node_to_refresh2] = [(S.name, 'R'), path[0] ]
        elif case1 and case2:
            start_end_node_update_storage[node_to_refresh1] = [(S.name, 'L'), node_to_refresh2 ]
            start_end_node_update_storage[node_to_refresh2] = [(S.name, 'R'), node_to_refresh1 ]
        else:
            start_end_node_update_storage[path[0]] = [(S.name, 'L'), path[-1] ]
            start_end_node_update_storage[path[-1]] = [(S.name, 'R'), path[0] ]

    return(start_scaf_index)

def GiveLinkConnection(Contigs, contig_objects1, contig_objects2, side1, side2, nr_links):
    if side1 == 'R' and side2 == 'L':
        max_pos = 0
        for contig in contig_objects1:
            if contig.position >= max_pos:
                linking_contig1 = contig
                max_pos = contig.position
        min_pos = sys.maxint
        for contig in contig_objects2:
            if contig.position <= min_pos:
                linking_contig2 = contig
                min_pos = contig.position

        linking_contig1.links[linking_contig2.name] = nr_links
        linking_contig2.links[linking_contig1.name] = nr_links


    elif side1 == 'L' and side2 == 'R':
        max_pos = 0
        for contig in contig_objects2:
            if contig.position >= max_pos:
                linking_contig2 = contig
                max_pos = contig.position

        min_pos = sys.maxint
        for contig in contig_objects1:
            if contig.position <= min_pos:
                linking_contig1 = contig
                min_pos = contig.position
        linking_contig1.links[linking_contig2.name] = nr_links
        linking_contig2.links[linking_contig1.name] = nr_links


    elif side1 == 'R' and side2 == 'R':
        max_pos = 0
        for contig in contig_objects1:
            if contig.position >= max_pos:
                linking_contig1 = contig
                max_pos = contig.position

        max_pos = 0
        for contig in contig_objects2:
            if contig.position >= max_pos:
                linking_contig2 = contig
                max_pos = contig.position
        linking_contig1.links[linking_contig2.name] = nr_links
        linking_contig2.links[linking_contig1.name] = nr_links


    elif side1 == 'L' and side2 == 'L':
        min_pos = sys.maxint
        for contig in contig_objects1:
            if contig.position <= min_pos:
                linking_contig1 = contig
                min_pos = contig.position
        min_pos = sys.maxint
        for contig in contig_objects2:
            if contig.position <= min_pos:
                linking_contig2 = contig
                min_pos = contig.position

        linking_contig1.links[linking_contig2.name] = nr_links
        linking_contig2.links[linking_contig1.name] = nr_links



'''
Created on Jun 17, 2013

@author: ksahlin
'''

import diploid

def main(contigs):
    kmer_dict, contig_dict = diploid.get_kmers(contigs)
    potential_haplotypes = diploid.get_haplotype_regions(kmer_dict, contig_dict)
    diploid.smith_waterman(potential_haplotypes)

    return




'''
Created on Jun 19, 2013

@author: ksahlin
'''

import os

import networkx as nx

from BESST import Contig, Scaffold
import smith_waterman

def reverse_complement(string):
    #rev_nuc = {'A':'T', 'C':'G', 'G':'C', 'T':'A', 'N':'N', 'X':'X'}
    # Modified for Abyss output
    rev_nuc = {'A':'T', 'C':'G', 'G':'C', 'T':'A', 'a':'t', 'c':'g', 'g':'c', 't':'a', 'N':'N', 'X':'X', 'n':'n', 'Y':'R', 'R':'Y', 'K':'M', 'M':'K', 'S':'S', 'W':'W', 'B':'V', 'V':'B', 'H':'D', 'D':'H', 'y':'r', 'r':'y', 'k':'m', 'm':'k', 's':'s', 'w':'w', 'b':'v', 'v':'b', 'h':'d', 'd':'h'}

    rev_comp = ''.join([rev_nuc[nucl] for nucl in reversed(string)])
    return(rev_comp)

def k_mer_graph_to_scaffold_graph(graph, removed):
    ''' We want to make new contigs of the haplotypic regions that we have solved.
        The regions that we have solved are the ones where a node has precisely 
        one neighbor in the overlapping k-mer graph.
    
    '''

    new_contigs = nx.Graph()

    for edge in graph.edges_iter():
        try:
            graph.edge[edge[0]][edge[1]]['m']
            node = edge[0]
            nbr = edge[1]
        except KeyError:
            continue
        ##
        # Modulo 2 operator to see what contig end we should map to
        # Beginning and forward of contig, or end and reverse compl. -> "Left" side of contig
        # Othervise right side

        if node[1] == 0:
            new_node1 = (node[0], 'L')
            other_side1 = (node[0], 'R')
        else:
            new_node1 = (node[0], 'R')
            other_side1 = (node[0], 'L')

        if nbr[1] == 0:
            new_node2 = (nbr[0], 'L')
            other_side2 = (nbr[0], 'R')
        else:
            new_node2 = (nbr[0], 'R')
            other_side2 = (nbr[0], 'L')
        new_contigs.add_edge(new_node1, new_node2)
        new_contigs.add_edge(new_node1 , other_side1)
        new_contigs.add_edge(new_node2 , other_side2)

    new_removed = {}
    for rem, used in removed.items():
        #if rem[1] == used[1] :
        new_removed[(rem[0], 'L')] = (used[0], 'L')
        new_removed[(rem[0], 'R')] = (used[0], 'R')
        #else:
        #    new_removed[(rem[0], 'L')] = (used[0], 'R')
        #    new_removed[(rem[0], 'R')] = (used[0], 'L')

    print len(new_contigs.nodes())
    return(new_contigs, new_removed)

def get_joined_sequence(contigs, sub_graph, start, end, k_mer_size, removed):
    sequence = ''
    print 'Nr nodes subgraph:', len(sub_graph.nodes())
    path = nx.algorithms.shortest_path(sub_graph, start, end)
    path_len = len(path)
    print 'Nr contigs in path:', path_len / 2.0

    index_to_remove = []
    for i, node in enumerate(path):
        if i % 2 == 0:
            try:
                contigs[node[0]]
            except KeyError:
                print 'Contig already removed. Number ' + str(i) + ' in path of length' + str(len(path))
                if i == 0 or i == len(path) - 2:# end contig. Just regular haplotype!
                    print 'End contig. Just regular haplotype! continue merging'
                    index_to_remove.append(i)
                    index_to_remove.append(i + 1)
                else:
                    print 'It is contig:', node[0]
                    return()
    for index in sorted(index_to_remove, reverse=True):
        #path[index] = removed[path[index]]
        print 'Trouble with dissappeared haplotype here!'
        print path[index], removed[path[index]]
        del path[index]

    path_len = len(path)
    for i, node in enumerate(path):
        if i % 2 == 1 and i < path_len - 2:
            ctg1 = contigs[node[0]] if node[1] == 'R' else reverse_complement(contigs[node[0]])
            seq1 = ctg1[:k_mer_size - 1]
            if i == 1:
                sequence += ctg1

            ctg2 = contigs[path[i + 1][0]] if path[i + 1][1] == 'L' else reverse_complement(contigs[path[i + 1][0]])
            seq2 = ctg2[:k_mer_size - 1]
            score, max_i, max_j = smith_waterman.SW(seq1, seq2, 0, 0)
            neg_gap = min(max_i, max_j)
            sequence += ctg2[neg_gap:]
            # Remove the contigs that are included in a new contig so
            # that we don't print it out several times
            del contigs[node[0]]

    return(sequence)

def merge_ctgs(new_contigs, removed, node):
    replacing_hapl = removed[node]
    other_side_hapl = (replacing_hapl[0], 'R') if replacing_hapl[1] == 'L' else (replacing_hapl[0], 'L')
    if len(new_contigs.neighbors(other_side_hapl)) > 1:
        print "warning!: Trying to merge haploid that are not ends of components"
        print other_side_hapl, new_contigs.neighbors(other_side_hapl)
        return()
    other_side = (node[0], 'R') if node[1] == 'L' else (node[0], 'L')
    nbrs = new_contigs.neighbors(other_side)
    nbr = filter(lambda x: x[0] != other_side[0], nbrs)
    if len(nbr) > 1:
        print "Warning: many neighbors for haplotype "
    nbr = nbr[0]
    new_contigs.remove_node((node[0], 'R'))
    new_contigs.remove_node((node[0], 'L'))
    new_contigs.add_edge(nbr, other_side_hapl)
    return()

def  get_remaining_overlaps(new_contigs, removed):

    for component in nx.connected_component_subgraphs(new_contigs):
        for node in component:
            if len(new_contigs.neighbors(node)) == 1:
                start = node
                break
        for node in component:
            if len(new_contigs.neighbors(node)) == 1 and node != start:
                end = node
                break
        if start in removed:
            merge_ctgs(new_contigs, removed, start)
        if end in removed:
            merge_ctgs(new_contigs, removed, end)

    return()


def print_out_contigs(contigs, new_contigs, output_dir, k_mer_size, removed):
    contigs_out = open(os.path.join(output_dir, 'BESST_contigs.fa'), 'w')

    contig_index = 0
    for comp in nx.connected_component_subgraphs(new_contigs):
        print len(comp.nodes()) / 2.0

    get_remaining_overlaps(new_contigs, removed)

    print 'After merge:'
    for comp in nx.connected_component_subgraphs(new_contigs):
        print len(comp.nodes()) / 2.0

    for component in nx.connected_component_subgraphs(new_contigs):
        contig_index += 1
        accession = '>BESST_ctg' + str(contig_index)

        for node in component:
            if len(new_contigs.neighbors(node)) == 1:
                start = node
                break
        for node in component:
            if len(new_contigs.neighbors(node)) == 1 and node != start:
                end = node
                break
        ctg_seq = get_joined_sequence(contigs, component, start, end, k_mer_size, removed)
        if ctg_seq:
            print >> contigs_out, accession + '\n' + ctg_seq




    # Print out remaining contigs (the ones that has not been 
    # removed as haplotypes or merged into bigger contigs)
    print len(contigs)
    cntru = 0
    for acc, sequence in contigs.items():
        cntru += 1
        print >> contigs_out, '>' + acc + '\n' + sequence
    print cntru, ' contigs that were kept as they are (not merged or removed).'
    return()

def generate_fasta(graph, contigs, output_dir, k_mer_size):

    haplotypes = open(os.path.join(output_dir, 'haplotypes.fa'), 'w')
    nodes_to_remove = set()
    haplotype_count = 0
    removed = {}
    for node in graph:
        if graph.node[node]['r']:
            nodes_to_remove.add(node)
            nbrs = graph.neighbors(node)
            for nbr in nbrs:
                if graph.edge[node][nbr].keys() == ['a']:
                    #print node
                    #print 'nbr!!:', nbr
                    removed[node] = nbr

    graph.remove_nodes_from(nodes_to_remove)
    for node in nodes_to_remove:
        # remove the nodes from the contig dictionary as well
        # so that we do not print them out among the new contigs.
        try:
            print >> haplotypes, '>' + node[0].strip() + '\n' + contigs[node[0]]
            del contigs[node[0]]
            haplotype_count += 1
        except KeyError:
            pass

    print 'Number of haplotypes removed:', haplotype_count
    print 'Nr allele nodes removed:', len(removed)
    new_contigs, removed = k_mer_graph_to_scaffold_graph(graph, removed)
    print_out_contigs(contigs, new_contigs, output_dir, k_mer_size, removed)

    return()



'''
    Created on Jun 17, 2012
    
    @author: ksahlin
    
    This file is part of BESST.
    
    BESST is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
    
    BESST is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    
    You should have received a copy of the GNU General Public License
    along with BESST.  If not, see <http://www.gnu.org/licenses/>.
    '''


import networkx as nx
from itertools import combinations
import random
import multiprocessing , Queue
import time
import ExtendLargeScaffolds as ELS


class Worker(multiprocessing.Process):

    def __init__(self, work_queue, result_queue):

        # base class initialization
        multiprocessing.Process.__init__(self)

        # job management stuff
        self.work_queue = work_queue
        self.result_queue = result_queue
        self.kill_received = False
    def run(self):
        #while not self.kill_received:

            # get a task
            #job = self.work_queue.get_nowait()
        #try:
        job = self.work_queue #.get_nowait()
        start_nodes = job[0]
        G_prime = job[1]
        #small_scaffolds = job[2]
        end = job[2]
        param = job[3]
            #all_paths_sorted_wrt_score = [(1,[]),(2,[])]

            # except Queue.Empty:
            #    print 'Noeee! empty!!', multiprocessing.current_process().name
            #    break

        print 'Enter ', multiprocessing.current_process().name
            # the actual processing
        all_paths_sorted_wrt_score = ELS.BetweenScaffolds(G_prime, end, start_nodes, param)
        #all_paths_sorted_wrt_score = [(1,[]),(2,[])]
            #nr_paths = find_all_paths(graph, start_node, end_node)
            # store the result
        self.result_queue.put(all_paths_sorted_wrt_score)
        print 'Exit', multiprocessing.current_process().name






'''
Created on Jun 26, 2013

@author: ksahlin
'''

import ctypes
import os

import BESST.diploid

class ReturnValues(ctypes.Structure):
    _fields_ = [("score", ctypes.c_int),
                ("i_max", ctypes.c_int),
                ("j_max", ctypes.c_int)]

def wrap_sw(seq1, seq2, mu, delta):
    c_s1 = ctypes.c_char_p(seq1)
    c_s2 = ctypes.c_char_p(seq2)
    res = ReturnValues(0, 0, 0)
    path = os.path.join( '/'.join(BESST.diploid.__file__.split('/')[:-1] ), 'lib/smith_waterman.so' )
    SW_fcn = ctypes.CDLL(path) #BESST/diploid/
    SW_fcn.SW(c_s1, c_s2, ctypes.byref(res)) # linux or when mingw used on windows
    return(res.score, res.i_max, res.j_max)

if __name__ == '__main__':
    wrap_sw('AGAAGGAAGGGGGAGAGTTTG',
            'AGAAGGAAGGTTGGGAGAGTTTG' , 0, 0)



'''
    Created on Mar 31, 2012

    @author: ksahlin

    This file is part of BESST.

    BESST is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    BESST is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with BESST.  If not, see <http://www.gnu.org/licenses/>.
    '''
import subprocess


class parameter(object):
    '''
    classdocs
    '''
    def __init__(self, parameter_mean_coverage=None, parameter_std_dev_coverage=None,
                 parameter_mean_ins_size=None , parameter_std_dev_ins_size=None,
                 parameter_output_directory=None, parameter_bamfile=None,
                 parameter_read_len=None, parameter_rel_weight=None,
                 parameter_ins_size_threshold=None, parameter_contigfile=None,
                 parameter_edgesupport=None, parameter_contig_threshold=None,
                 parameter_scaffold_indexer=0, parameter_first_lib=None,
                 parameter_cov_cutoff=None, parameter_tot_assembly_length=None,
                 parameter_current_N50=None, parameter_current_L50=None,
                 parameter_hapl_ratio=None, parameter_hapl_threshold=None,
                 parameter_detect_haplotype=None, parameter_detect_duplicate=None,
                 parameter_gff_file=None, parameter_information_file=None,
                 parameter_fosmidpool=None, parameter_extend_paths=None,
                 parameter_development=None, parameter_plots=None, parameter_path_threshold=None,
                 path_gaps_estimated =0, parameter_gap_estimations = [],
                 contamination_mean=None, contamination_stddev = None, contamination_ratio=0,
                 no_score=None):

        # Library information
        # Contig information
        # Algorithm parameters
        # Assembly information
        # Algorithm information
        # Output information
        self.mean_coverage = parameter_mean_coverage
        self.std_dev_coverage = parameter_std_dev_coverage
        self.mean_ins_size = parameter_mean_ins_size
        self.std_dev_ins_size = parameter_std_dev_ins_size
        self.output_directory = parameter_output_directory
        self.bamfile = parameter_bamfile
        self.read_len = parameter_read_len
        self.rel_weight = parameter_rel_weight
        self.ins_size_threshold = parameter_ins_size_threshold
        self.contigfile = parameter_contigfile
        self.edgesupport = parameter_edgesupport
        self.contig_threshold = parameter_contig_threshold
        self.scaffold_indexer = parameter_scaffold_indexer
        self.first_lib = parameter_first_lib
        self.cov_cutoff = parameter_cov_cutoff
        self.tot_assembly_length = parameter_tot_assembly_length
        self.current_N50 = parameter_current_N50
        self.current_L50 = parameter_current_L50
        self.hapl_ratio = parameter_hapl_ratio
        self.hapl_threshold = parameter_hapl_threshold
        self.detect_haplotype = parameter_detect_haplotype
        self.detect_duplicate = parameter_detect_duplicate
        self.gff_file = parameter_gff_file
        self.information_file = parameter_information_file
        self.fosmidpool = parameter_fosmidpool
        self.extend_paths = parameter_extend_paths
        self.development = parameter_development
        self.plots = parameter_plots
        self.path_threshold = parameter_path_threshold
        self.no_score = no_score

        # related to gap distances
        self.path_gaps_estimated = path_gaps_estimated
        self.gap_estimations = parameter_gap_estimations

        # If MP library (PE contamination)
        self.contamination_ratio = contamination_ratio
        self.contamination_mean = contamination_mean
        self.contamination_stddev = contamination_stddev

class counters(object):
    def __init__(self, param_count=None, param_non_unique=None, param_non_unique_for_scaf=None,
                 param_nr_of_duplicates=None, param_prev_obs1=None, param_prev_obs2=None,
                 param_reads_with_too_long_insert=None):

        self.count = param_count
        self.non_unique = param_non_unique
        self.non_unique_for_scaf = param_non_unique_for_scaf
        self.nr_of_duplicates = param_nr_of_duplicates
        self.prev_obs1 = param_prev_obs1
        self.prev_obs2 = param_prev_obs2
        self.reads_with_too_long_insert = param_reads_with_too_long_insert

#
#def MemoryUsage():
#    percent_usage = subprocess.Popen("ps -v | grep 'python Main.py' | awk '{sum+=$12} END {print sum}'",
#                                     shell=True,
#                                     stdout=subprocess.PIPE,
#                                     )
#    stdout_list = percent_usage.communicate()[0]
#    print 'Percentage of memory occupied by BESST from total memory: ', stdout_list
#    return()







import os
from setuptools import setup, find_packages

setup(
    name='BESST',
    version='1.1.2',
    packages=find_packages(exclude=['ez_setup', 'examples', 'tests']),
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: Science/Research",
        "Natural Language :: English",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python",
        "Topic :: Scientific/Engineering :: Bio-Informatics"
      ],
    scripts=['runBESST', 'docs/INSTALL.txt', 'docs/MANUAL.txt', 'scripts/reads_to_ctg_map.py'],
    description='Scaffolder for genomic assemblies.',
    author='Kristoffer Sahlin',
    author_email='kristoffer.sahlin@scilifelab.se',
    url='https://github.com/ksahlin/BESST',
    license='GPLv3',
    long_description=open(os.path.join(os.getcwdu(), 'README.md')).read(),
    install_requires=['pysam==0.7',
                      'networkx>=1.4',
                      'mathstats>=0.1.2',
                      'scipy>=0.9',
                      'pulp>=1.4'],
)



'''
    Created on Oct 30, 2012
    
    @author: ksahlin
    
    This file is part of BESST.
    
    BESST is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
    
    BESST is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    
    You should have received a copy of the GNU General Public License
    along with BESST.  If not, see <http://www.gnu.org/licenses/>.
    '''


import multiprocessing, Queue
import ExtendLargeScaffolds as ELS
class Worker(multiprocessing.Process):

    def __init__(self, work_queue, result_queue):

        # base class initialization
        multiprocessing.Process.__init__(self)

        # job management stuff
        self.work_queue = work_queue
        self.result_queue = result_queue
        self.kill_received = False
    def run(self):
        while not self.kill_received:

                # get a task
                #job = self.work_queue.get_nowait()
            worker_name = multiprocessing.current_process().name
            try:
                jobs, G, G_prime, small_scaffolds, end = self.work_queue.get_nowait()
            except Queue.Empty:
                break

            # the actual processing
            all_paths_sorted_wrt_score = ELS.BetweenScaffolds(G, G_prime, small_scaffolds, end, jobs)
            self.result_queue.put(all_paths_sorted_wrt_score)
            print 'Exited from ', worker_name

        #return()

#            for job in jobs:
#                start_node = job[0]
#                end_node = job[1]
#                #nr_paths = find_all_paths(graph, start_node, end_node)
#                    #print nr_paths, multiprocessing.current_process().name
#                # store the result
#                self.result_queue.put(nr_paths)
#import os

    #result_queue.put(all_paths_sorted_wrt_score)

    #os.kill()

result_list = []
def log_result(result):
    # This is called whenever foo_pool(i) returns a result.
    # result_list is modified only by the main process, not the pool workers.
    print 'Going in here'
    result_list.append(result)
def worker(jobs, G, G_prime, small_scaffolds, end):
    worker_name = multiprocessing.current_process().name
    # the actual processing
    all_paths_sorted_wrt_score = ELS.BetweenScaffolds(G, G_prime, small_scaffolds, end, jobs)
    print 'Exited from ', worker_name
    return(all_paths_sorted_wrt_score)

def BetweenScaffoldsParallellized(G, G_prime, small_scaffolds, num_processes, end):
    import multiprocessing , Queue
    #import worker #(multi processing)
    import heapq

    #import random
#    end = set()
#    for node in G:
#        end.add(node)

    # load up work queue
    work_queue = multiprocessing.Queue()
    nodes = G.nodes()
    #random.shuffle(nodes)
    nr_jobs = len(nodes)
    chunk = nr_jobs / num_processes
    #print jobs, nr_jobs, chunk
    counter = 0

    # partition equally many nodes in G to each core
    while counter < nr_jobs:
        work_queue.put((set(nodes[counter:counter + chunk]), G, G_prime, small_scaffolds, end))
        print 'node nr', counter, 'to', counter + chunk - 1, 'added'
        counter += chunk

    # create a queue to pass to workers to store the results
    #result_queue = multiprocessing.Queue()    

    pool = multiprocessing.Pool()
    #num_processes = multiprocessing.cpu_count()
    #for i in range(0,num_processes):
    while not work_queue.empty():
        nodes, G, G_prime, small_scaffolds, end = work_queue.get_nowait()
        pool.apply_async(worker, args=(nodes, G, G_prime, small_scaffolds, end), callback=log_result)
    pool.close()
    pool.join()
    #print(result_list)

#
#    jobs = []
#    for i in range(num_processes):
#        try:
#            nodes, G, G_prime, small_scaffolds, end = work_queue.get_nowait()
#            print 'Time to work'
#        except Queue.Empty:
#            print 'lol no work'
#        p = multiprocessing.Process(target=worker.worker, args=(nodes, G, G_prime, small_scaffolds, end,result_queue))
#        jobs.append(p)
#        print 'job nr ',i, 'start'
#        p.start()

#    
#    # spawn workers (equally many as available cores)
#    workers =[]
#    for i in range(0,num_processes):
#        workers.append(MP.Worker(work_queue, result_queue))
#        workers[i].start()
    #print 'hejhej!'
    work_queue.close() # Indicate that no more data will be put on this queue by the current process
    # Wait for the workers to finish
#    for i in range(0,num_processes):
#        jobs[i].join()
#        print "I'm waiting..."
#        #workers[i].join()

    # collect the results off the queue
    def wrapper(func, args):
        return(func(*args))
    #tot_result = []

    #while not result_queue.empty():
    #    tot_result.append(result_queue.get())
    #print  'tot result', tot_result
    #Merge the scored paths from each of the subprocesses into one consensus path
    itr = wrapper(heapq.merge, result_list) #tot_result)

    return(itr)



'''
Created on Aug 12, 2013

@author: ksahlin
'''

import os

try:
    import matplotlib
    matplotlib.use('Agg')
    import matplotlib.pyplot as plt
except ImportError, RuntimeError:
    pass


def histogram(x_, param, bins=20, x_label='x', y_label='y', title='Histogram',nr_obs = 10000):
    x = x_[:nr_obs] # If many contigs/edges we only plot 10 000 (default) for time and visuability purposes
    dest = os.path.join(param.output_directory, 'plots', title + '.png')
    plt.hist(x, bins)
    plt.xlabel(x_label)
    plt.ylabel(y_label)
    plt.title(title)

    plt.savefig(dest)
    plt.clf()
    return()

def dot_plot(x_, y_, param, x_label='x', y_label='y', title='Dotplot'):
    x = x_[:10000] # If many contigs/edges we only plot 10 000 for time and visuability purposes
    y = y_[:10000] # If many contigs/edges we only plot 10 000 for time and visuability purposes
    dest = os.path.join(param.output_directory, 'plots', title + '.png')
    plt.scatter(x, y, marker='o')
    plt.xlabel(x_label)
    plt.ylabel(y_label)
    plt.title(title)
    plt.grid(True)

    plt.savefig(dest)
    plt.clf()
    return


def multiple_histogram(list_of_datasets_, param, x_label='x', y_label='y', title='Stacked_histogram'):
    list_of_datasets = [list[:10000] for list in list_of_datasets_]
    dest = os.path.join(param.output_directory, 'plots', title + '.png')
    # filter out if any of the lists contains 0 elemnets
    list_of_datasets = filter(lambda x: len(x) > 0, list_of_datasets)
    for dataset in list_of_datasets:
        plt.hist(dataset, alpha=0.5)
    plt.xlabel(x_label)
    plt.ylabel(y_label)
    plt.title(title)

    plt.savefig(dest)
    plt.clf()

    return()





#def VizualizeGraph(G, param, Information):
#    import os
#    try:
#        import matplotlib
#        matplotlib.use('Agg')
#
#        try:
#            os.mkdir(param.output_directory + '/graph_regions' + str(int(param.mean_ins_size)))
#        except OSError:
#            #directory is already created
#            pass
#        counter = 1
#        import copy
#
#        G_copy = copy.deepcopy(G)
#        RemoveIsolatedContigs(G_copy, Information)
#        CB = nx.connected_component_subgraphs(G_copy)
#        for cycle in CB:
#            nx.draw(cycle)
#            matplotlib.pyplot.savefig(param.output_directory + 'graph_regions' + str(int(param.mean_ins_size)) + '/' + str(counter) + '.png')
#            matplotlib.pyplot.clf()
#            counter += 1
#    except ImportError:
#        pass
#    return()



'''
    Created on May 30, 2014

    @author: ksahlin

    This file is part of BESST.

    BESST is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    BESST is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with BESST.  If not, see <http://www.gnu.org/licenses/>.
    '''

import copy
import math
import random
from collections import Counter

from pulp import *
from mathstats.normaldist.normal import normpdf
from mathstats.normaldist.truncatedskewed import param_est as GC




class Contig(object):
    """Container with contig information for a contig in a path"""
    def __init__(self, index, length):
        super(Contig, self).__init__()
        self.index = index
        self.length = length
        self.position = None
        

class Path(object):
    """Contains all information of a path. This is basically a supgraph of the 
    Scaffold graph in BESST contianing all contigs in a path that has high score 
    and is going to be made into a scaffold. 

    Path is an object with methods for calculating likelihood of a path given link observations.
    Due to computational requiremants, we don't calculate the true likelihoos all paths usually have thousands
    of links. Instead, we take the average link obervation between all contigs. This will not give true ML 
    estimates but speeds up calculation with thousands of x order. In practice, using average link obervation 
    For ML estimation will give a fairly good prediction. For this cheat, see comment approx 15 lines below.  """
    def __init__(self, ctg_lengths, observations, param, initial_path=False):
        super(Path, self).__init__()
        self.mean = param.mean_ins_size
        self.stddev = param.std_dev_ins_size
        self.read_len = param.read_len
        self.contamination_ratio = param.contamination_ratio
        self.contamination_mean = param.contamination_mean
        self.contamination_stddev = param.contamination_stddev
        self.ctgs = []
        for i,length in enumerate(ctg_lengths):
            self.ctgs.append(Contig(i, length))
        self.ctgs = tuple(self.ctgs)
        self.gaps = [0]*(len(ctg_lengths)-1) # n contigs has n-1 gaps between them, start with gap size 0  
        
        # get positions for when all gaps are 0
        self.update_positions()

        # let us cheat here! Instead of calculating likeliooods of thousands of
        # onservations we calculate the ikelihood for them average (mean) of the
        # observations and weight it with the number of observations
        obs_dict = {}
        for c1,c2,is_PE_link in observations:
            nr_obs = len(observations[(c1,c2,is_PE_link)])
            if is_PE_link:
                PE_obs = map(lambda x: self.ctgs[c1].length + self.ctgs[c2].length - x + 2*param.read_len ,observations[(c1,c2,is_PE_link)])
                mean_obs = sum( PE_obs)/nr_obs
                obs_dict[(c1,c2,is_PE_link)] = (mean_obs,nr_obs)
                if mean_obs > self.contamination_mean + 6 * self.contamination_stddev and not initial_path:
                    self.observations = None
                    return None
            else:
                mean_obs = sum(observations[(c1,c2,is_PE_link)])/nr_obs
                obs_dict[(c1,c2,is_PE_link)] = (mean_obs,nr_obs)
            

        self.observations = obs_dict
        #print self.observations

        # for c1,c2 in self.observations:
        #     if self.observations[(c1,c2)][0] > 1500:
        #         print self.observations


    def get_distance(self,start_index,stop_index):
        total_contig_length = sum(map(lambda x: x.length ,filter(lambda x: start_index <= x.index < stop_index, self.ctgs) ))
        total_gap_length = sum(self.gaps[start_index:stop_index])
        index_adjusting = len(self.gaps[start_index:stop_index]) # one extra bp shifted each time
        return (total_contig_length, total_gap_length, index_adjusting)

    def update_positions(self):
        for ctg in self.ctgs:
            index = ctg.index
            ctg.position = sum(self.get_distance(0,index))

    def get_inferred_isizes(self):
        self.isizes = {}

        for (c1,c2,is_PE_link) in self.observations:
            gap = self.ctgs[c2].position - (self.ctgs[c1].position + self.ctgs[c1].length) - (c2-c1) # last thing is an index thing
            #x = map(lambda obs: obs[0] + gap , self.observations[(c1,c2)]) # inferr isizes
            x = self.observations[(c1,c2,is_PE_link)][0] + gap
            self.isizes[(c1,c2,is_PE_link)] = x

    def get_GapEst_isizes(self):
        self.gapest_predictions = {}
        for (i,j,is_PE_link) in self.observations:
            mean_obs = self.observations[(i,j,is_PE_link)][0]
            if is_PE_link:
                self.gapest_predictions[(i,j,is_PE_link)] = self.observations[(i,j,is_PE_link)][0] + GC.GapEstimator(self.contamination_mean, self.contamination_stddev, self.read_len, mean_obs, self.ctgs[i].length, self.ctgs[j].length)
            else:
                self.gapest_predictions[(i,j,is_PE_link)] = self.observations[(i,j,is_PE_link)][0] + GC.GapEstimator(self.mean, self.stddev, self.read_len, mean_obs, self.ctgs[i].length, self.ctgs[j].length)

            #print mean_obs, self.ctgs[i].length, self.ctgs[j].length, 'gap:' ,  GC.GapEstimator(self.mean, self.stddev, self.read_len, mean_obs, self.ctgs[i].length, self.ctgs[j].length)
        


    def new_state_for_ordered_search(self,start_contig,stop_contig,mean,stddev):
        """
            This function gives a new state between two contigs c1 and c2 in the contig path
            that we will change the gap size for.
            Currently, we change the gap to expected_mean_obs - mean_observation, e.g. a
            semi-naive estimation. More variablity in gap prediction could be acheved in the 
            same way as described for function propose_new_state_MCMC().
             
        """
        new_path = copy.deepcopy(self) # create a new state
        (c1,c2) = (start_contig,stop_contig)
        mean_obs = self.observations[(c1,c2,0)][0] if (c1,c2,0) in self.observations else self.observations[(c1,c2,1)][0]  # take out observations and
        exp_mean_over_bp = mean + stddev**2/float(mean+1)
        proposed_distance = exp_mean_over_bp - mean_obs # choose what value to set between c1 and c2 

        #print 'CHOSEN:', (c1,c2), 'mean_obs:', mean_obs, 'proposed distance:', proposed_distance
        (total_contig_length, total_gap_length, index_adjusting) = self.get_distance(c1+1,c2)
        #print 'total ctg length, gap_lenght,index adjust', (total_contig_length, total_gap_length, index_adjusting)
        avg_suggested_gap = (proposed_distance - total_contig_length) / (c2-c1)
        #print avg_suggested_gap, proposed_distance, total_contig_length, c2-c1
        for index in range(c1,c2):
            new_path.gaps[index] = avg_suggested_gap
        #new_path.gaps[index] = proposed_distance
        return new_path


    # def calc_log_likelihood(self,mean,stddev):
    #     log_likelihood_value = 0
    #     exp_mean_over_bp = mean + stddev**2/float(mean+1)
    #     for (c1,c2) in self.isizes:
    #         log_likelihood_value += math.log( normpdf(self.isizes[(c1,c2)],exp_mean_over_bp, stddev) ) * self.observations[(c1,c2)][1]
    #         #for isize in self.isizes[(c1,c2)]:
    #         #    log_likelihood_value += math.log( normpdf(isize,mean,stddev) )
            
    #     return log_likelihood_value

    def calc_dist_objective(self):
        objective_value = 0
        self.get_GapEst_isizes()
        for (c1,c2, is_PE_link) in self.isizes:
            objective_value += abs(self.gapest_predictions[(c1,c2,is_PE_link)] - self.isizes[(c1,c2,is_PE_link)]) * self.observations[(c1,c2,is_PE_link)][1]
            #for isize in self.isizes[(c1,c2)]:
            #    objective_value += math.log( normpdf(isize,mean,stddev) )
            
        return objective_value

    def make_path_dict_for_besst(self):
        path_dict = {}
        for ctg1,ctg2 in zip(self.ctgs[:-1],self.ctgs[1:]):
            path_dict[(ctg1,ctg2)] = ctg2.position - (ctg1.position + ctg1.length) - 1 
        #print path_dict
        return path_dict


    def calc_probability_of_LP_solution(self, help_variables):
        log_prob = 0
        for (i,j,is_PE_link),variable in help_variables.iteritems():
            print self.contamination_ratio * self.observations[(i,j,is_PE_link)][1] * normpdf(variable.varValue,self.contamination_mean,self.contamination_stddev)
            if is_PE_link:
                try:
                    log_prob += math.log(self.contamination_ratio * self.observations[(i,j,is_PE_link)][1] * normpdf(variable.varValue,self.contamination_mean,self.contamination_stddev)) 
                except ValueError:
                    log_prob += - float("inf")
            else:
                try:
                    log_prob += math.log((1 - self.contamination_ratio) * self.observations[(i,j,is_PE_link)][1]* normpdf(variable.varValue,self.contamination_mean,self.contamination_stddev)) 
                except ValueError:
                    log_prob += - float("inf")
        return log_prob

    def LP_solve_gaps(self):
        exp_means_gapest = {}

        for (i,j,is_PE_link) in self.observations:
            mean_obs = self.observations[(i,j,is_PE_link)][0]
            if is_PE_link:
                exp_means_gapest[(i,j,is_PE_link)] = self.observations[(i,j,is_PE_link)][0] + GC.GapEstimator(self.contamination_mean, self.contamination_stddev, self.read_len, mean_obs, self.ctgs[i].length, self.ctgs[j].length)
                #print 'GAPEST:',mean_obs, self.ctgs[i].length, self.ctgs[j].length, 'gap:' ,  GC.GapEstimator(self.contamination_mean, self.contamination_stddev, self.read_len, mean_obs, self.ctgs[i].length, self.ctgs[j].length)

            else:
                exp_means_gapest[(i,j,is_PE_link)] = self.observations[(i,j,is_PE_link)][0] + GC.GapEstimator(self.mean, self.stddev, self.read_len, mean_obs, self.ctgs[i].length, self.ctgs[j].length)
                #print 'GAPEST:',mean_obs, self.ctgs[i].length, self.ctgs[j].length, 'gap:' ,  GC.GapEstimator(self.mean, self.stddev, self.read_len, mean_obs, self.ctgs[i].length, self.ctgs[j].length)
        

        #exp_mean_over_bp = self.mean + self.stddev**2/float(self.mean+1)

        #calculate individual exp_mean over each edge given observation with gapest??

        gap_vars= []
        for i in range(len(self.ctgs)-1):
            gap_vars.append( LpVariable(str(i), None, self.mean + 2*self.stddev, cat='Integer'))

        # help variables because objective function is an absolute value
        help_variables = {}
        for (i,j,is_PE_link) in self.observations:
            help_variables[(i,j,is_PE_link)] = LpVariable("z_"+str(i)+'_'+str(j)+'_'+str(is_PE_link), None, None,cat='Integer')

        # # variables to penalize negative gaps
        # penalize_variables = {}
        # for i in range(len(self.gaps)):
        #     penalize_variables[i] =  LpVariable("r_"+str(i)+'_'+str(j), None, 0,cat='Integer')

        # PENALIZE_CONSTANT = Counter()
        # for (i,j) in self.observations:
        #     for k in range(i,j-1): # all gaps between contig i and j
        #         # we penalize with the count of all observations spanning over the gap
        #         PENALIZE_CONSTANT[k] += self.observations[(i,j)][1] # number of observations spanning over the gap 

        problem = LpProblem("PathProblem",LpMinimize)

        #problem += lpSum( [ help_variables[(i,j,is_PE_link)]*self.observations[(i,j,is_PE_link)][1] for (i,j,is_PE_link) in self.observations] ) , "objective"
        if self.contamination_ratio:
            problem += lpSum( [ is_PE_link * (1 - self.contamination_ratio) * help_variables[(i,j,is_PE_link)]*self.observations[(i,j,is_PE_link)][1] + (1-is_PE_link)*(self.contamination_ratio)* help_variables[(i,j,is_PE_link)]*self.observations[(i,j,is_PE_link)][1] for (i,j,is_PE_link) in self.observations] ) , "objective"
        else:
            problem += lpSum( [ help_variables[(i,j,is_PE_link)]*self.observations[(i,j,is_PE_link)][1] for (i,j,is_PE_link) in self.observations] ) , "objective"

        # problem += lpSum( [ - penalize_variables[i]*PENALIZE_CONSTANT[i] for i in range(len(self.gaps))] )

        # adding constraints induced by the absolute value of objective function
        for (i,j,is_PE_link) in self.observations:
            problem += exp_means_gapest[(i,j,is_PE_link)] - sum(map(lambda x: x.length, self.ctgs[i+1:j])) - self.observations[(i,j,is_PE_link)][0] - lpSum( gap_vars[i:j] )  <= help_variables[(i,j,is_PE_link)] ,  "helpcontraint_"+str(i)+'_'+ str(j)+'_'+str(is_PE_link)

        for (i,j,is_PE_link) in self.observations:
            problem += - exp_means_gapest[(i,j,is_PE_link)] + lpSum( gap_vars[i:j] ) + sum(map(lambda x: x.length, self.ctgs[i+1:j])) + self.observations[(i,j,is_PE_link)][0]  <= help_variables[(i,j,is_PE_link)] ,  "helpcontraint_negative_"+str(i)+'_'+ str(j)+'_'+str(is_PE_link)


        # adding distance constraints

        # for (i,j,is_PE_link) in self.observations:
        #     # print 'order:',(i,j,is_PE_link)
        #     if is_PE_link:
        #         problem += lpSum( gap_vars[i:j] ) + sum(map(lambda x: x.length, self.ctgs[i+1:j])) + self.observations[(i,j,is_PE_link)][0]  <= self.contamination_mean +4*self.contamination_stddev,  "dist_constraint_"+str(i)+'_'+ str(j)+'_'+str(is_PE_link)                
        #     else:
        #         problem += lpSum( gap_vars[i:j] ) + sum(map(lambda x: x.length, self.ctgs[i+1:j])) + self.observations[(i,j,is_PE_link)][0]  <= self.mean +4*self.stddev,  "dist_constraint_"+str(i)+'_'+ str(j)+'_'+str(is_PE_link)

        # for (i,j,is_PE_link) in self.observations:
        #     if is_PE_link:
        #         problem += - lpSum( gap_vars[i:j] ) - sum(map(lambda x: x.length, self.ctgs[i+1:j])) - self.observations[(i,j,is_PE_link)][0]  <= - self.contamination_mean +4*self.contamination_stddev,  "dist_constraint_negative_"+str(i)+'_'+ str(j)+'_'+str(is_PE_link)
        #     else:
        #         problem += - lpSum( gap_vars[i:j] ) - sum(map(lambda x: x.length, self.ctgs[i+1:j])) - self.observations[(i,j,is_PE_link)][0]  <= - self.mean +4*self.stddev,  "dist_constraint_negative_"+str(i)+'_'+ str(j)+'_'+str(is_PE_link)

        # # Adding constraints induced from introducing a negative gap penalizer
        # for i in range(len(self.gaps)):
        #     problem += penalize_variables[i] - gap_vars[i]  <= 0 ,  "neg_gap_contraint_"+str(i)


        try:
            problem.solve()
        except : #PulpSolverError:
            problem.solve()
            print 'Could not solve LP, printing instance:'
            print 'Objective:'
            print problem.objective
            print 'Constraints:'
            print problem.constraints

            print 'Solving with ordered_search instead'
            path = ordered_search(self)
            return path.gaps

        optimal_gap_solution = [0]*(len(self.ctgs) -1)
        for v in problem.variables():
            try:
                optimal_gap_solution[int( v.name)] = v.varValue
                #print v.name, "=", v.varValue,
            except ValueError:
                #print v.name, "=", v.varValue,
                pass

        self.objective = value(problem.objective)
        #prob = self.calc_probability_of_LP_solution(help_variables)

        #print self.objective
        #print optimal_gap_solution
        return optimal_gap_solution

    def __str__(self):
        string= ''
        for ctg in self.ctgs:
            string += 'c'+str(ctg.index) +',startpos:'+ str(ctg.position)+',endpos:'+str(ctg.position + ctg.length)+'\n'

        return string

        
def ordered_search(path):

    path.get_inferred_isizes()

    for c1 in range(len(path.ctgs)-1):
        for c2 in range(c1+1,len(path.ctgs)):
            if (c1,c2) in path.observations:
                suggested_path = path.new_state_for_ordered_search(c1,c2,path.mean,path.stddev)
                suggested_path.update_positions()
                suggested_path.get_inferred_isizes()
                suggested_path.calc_dist_objective()
                if suggested_path.calc_dist_objective() < path.calc_dist_objective():
                    #print "SWITCHED PATH TO SUGGESTED PATH!!"
                    path = suggested_path
                    #print path
                else:
                    pass
                    #print 'PATH not taken!'
            else:
                continue
      
    # print 'FINAL PATH:'
    # print path
    # print 'With likelihood: ', path.calc_log_likelihood(mean,stddev)  

    return path

def main(contig_lenghts, observations, param, initial_path):
    """
    contig_lenghts: Ordered list of integers which is contig lengths (ordered as contigs comes in the path)
    observations:  dictionary oflist of observations, eg for contigs c1,c2,c3
                    we can have [(c1,c2):[23, 33, 21],(c1,c3):[12,14,11],(c2,c3):[11,34,32]]
    """

    path = Path(contig_lenghts,observations, param, initial_path=initial_path)
    
    if path.observations == None:
        return None

    # ML_path = MCMC(path,mean,stddev)
    # ML_path = ordered_search(path,mean,stddev)

    optimal_LP_gaps = path.LP_solve_gaps()
    path.gaps = optimal_LP_gaps
    path.update_positions()

    # print 'MCMC path:'
    # print ML_path
    # print 'MCMC path likelihood:',ML_path.calc_log_likelihood(mean,stddev)
    # print 'Ordered search path:'
    # print ML_path
    # print 'Ordered search likelihood:',ML_path.calc_log_likelihood(mean,stddev)
    # return ML_path

    return path

if __name__ == '__main__':
    contig_lenghts = [3000,500,500,500,3000]
    observations_normal = {(0,1):[1800,2000], (0,2):[1500,1800,1400,1700], (0,3):[1200,800,1000], 
                    (1,2):[750,800],(1,3):[600,700], (1,4):[300,600,700],
                    (2,3):[700,750], (2,4):[1400,1570],
                    (3,4):[2000,1750], }

    observations_linear = {(0,1):[450,500],  (1,2):[750,800],
                    (2,3):[700,750], (3,4):[400,170]}

    # The imortance of support:
    # one edge with a lot of observatins contradicting the others
    observations_matter = {(0,1):[450,500],  (1,2):[750,800],
                    (2,3):[700,750], (3,4):[400,170],
                    (0,4):[700]*30 }

    # long contig, 500bp gap, 3*short_contigs, long contig
    observations = {(0,1):[1450,1300,1200,1570], (0,2):[700,800,1000], (0,3):[250,300],
                    (1,2):[900],(1,3):[900,800], (1,4):[800,900,1000],
                    (2,3):[900], (2,4):[900,800],
                    (3,4):[1500,1750,1350,1900,1950] }

    # negative gaps test case here:

 
    mean = 1500
    stddev = 500
    read_len = 100
    main(contig_lenghts,observations,mean,stddev, read_len)


